#include "VisMonitoring/DQMServer/interface/VisDQMImageServer.h"
#include "VisMonitoring/DQMServer/interface/VisDQMRenderDaemon.h"
#include "TObjString.h"
#include "TDirectory.h"
#include "TFile.h"
#include "TKey.h"
#include "TH1.h"
#include <sys/time.h>

using namespace lat;

class VisDQMArchiveImageServer : public DQMImplNet<VisDQMObject>,
				 public VisDQMImageServer
{
  typedef std::map<std::string, std::pair<Time, TFile *> > FileMap;
  FileMap files_;
  size_t maxfiles_;

public:
  static const uint32_t DQM_MSG_GET_IMAGE_FILE	= 5;
  static const uint32_t DQM_REPLY_IMAGE_FILE	= 106;

  VisDQMArchiveImageServer(const ServerInfo &info, size_t maxfiles)
    : DQMImplNet<VisDQMObject>(info.argv[0]),
      VisDQMImageServer(info),
      maxfiles_(maxfiles)
    {
      logme()
	<< "INFO: DQM archive image server started"
	<< " in path '" << info_.statedir
	<< "' and at request port " << info_.port
	<< ", black list in " << info_.blacklistFile << std::endl;

      // Establish the server side.
      debug(info_.debug);
      delay(100);
      sendScalarAsText(true);
      startLocalServer(info_.port);
    }

  ~VisDQMArchiveImageServer(void)
    {
      for (FileMap::iterator i = files_.begin(), e = files_.end(); i != e; ++i)
      {
	i->second.second->Close();
	delete i->second.second;
      }
    }

protected:
  // Return the logging prefix.
  virtual std::ostream &
  logme(void)
    {
      return DQMNet::logme();
    }

  // Check if the server should stop.
  virtual bool
  shouldStop(void)
    {
      return info_.stopfile.exists();
    }

  // Purge dead objects.  This cannot ever happen for this server
  // because we are never connected to any network data source.
  virtual void
  purgeDeadObject(VisDQMObject &o)
    {
      ASSERT(false);
    }
  
  // Release objects from wait.  This cannot ever happen because this
  // server is never connected to a network data source which it would
  // have to wait for.
  void
  releaseFromWait(Bucket *msg, WaitObject &w, Object *o)
    {
      ASSERT(false);
    }

  // Web server facing message handling.  Respond to image requests.
  bool
  onMessage(Bucket *msg, Peer *p, unsigned char *data, size_t len)
    {
      // Decode and process this message.
      uint32_t type;
      memcpy (&type, data + sizeof(uint32_t), sizeof (type));
      if (type == DQM_MSG_GET_IMAGE_FILE)
      {
	if (debug_)
	  logme()
	    << "DEBUG: received message 'GET IMAGE FILE' from peer "
	    << p->peeraddr << std::endl;

	if (len < 5*sizeof(uint32_t))
	{
	  logme()
	    << "ERROR: corrupt 'GET IMAGE FILE' message of length " << len
	    << " from peer " << p->peeraddr << std::endl;
	  return false;
	}

	uint32_t namelen;
	uint32_t speclen;
	uint32_t filelen;
	memcpy(&namelen, data + 2*sizeof(uint32_t), sizeof(namelen));
	memcpy(&speclen, data + 3*sizeof(uint32_t), sizeof(speclen));
	memcpy(&filelen, data + 4*sizeof(uint32_t), sizeof(filelen));
	if (len != 5*sizeof(uint32_t) + namelen + speclen + filelen)
	{
	  logme()
	    << "ERROR: corrupt 'GET IMAGE FILE' message of length " << len
	    << " from peer " << p->peeraddr
	    << ", expected length " << (5*sizeof(uint32_t))
	    << " + " << namelen << " + " << speclen << " + " << filelen
	    << std::endl;
	  return false;
	}

	std::string name((char *) data + 5*sizeof(uint32_t), namelen);
	std::string spec((char *) data + 5*sizeof(uint32_t) + namelen, speclen);
	std::string file((char *) data + 5*sizeof(uint32_t) + namelen + speclen, filelen);

        // Give myself 30 seconds, then die.  ROOT sometimes gets stuck
        // here in an infinite loop when closing files, presumably due
        // to an occasional rampage around in memory while rendering.
        struct itimerval timer;
        timer.it_value.tv_sec = 30;
        timer.it_value.tv_usec = 0;
        timer.it_interval.tv_sec = 0;
        timer.it_interval.tv_usec = 0;
        setitimer(ITIMER_VIRTUAL, &timer, 0);

	// FIXME: Image caching? We lose the object each time around here.
	TObjString str;
	VisDQMObject o;
	o.version = 0;
	o.name = name;
	o.object = 0;
	o.reference = 0;
	o.flags = DQM_FLAG_DEAD;
	findFileObject(o, str, name, file);

        // Let timer go now, we made it through the file management,
	// and the rendering has its own internal suicide timer.
        timer.it_value.tv_sec = 0;
        timer.it_value.tv_usec = 0;
        timer.it_interval.tv_sec = 0;
        timer.it_interval.tv_usec = 0;
        setitimer(ITIMER_VIRTUAL, &timer, 0);

	// Now render.
	sendImageToPeer(msg, p, o, spec, file);

	// Release the objects.
	delete o.object;
	delete o.reference;
	return true;
      }
      else
      {
	logme()
	  << "ERROR: unrecognised message of length " << len
	  << " and type " << type << " from peer " << p->peeraddr
	  << std::endl;
	return false;
      }
  }

  // Send an image of an object to a peer.
  virtual void
  sendImageToPeer(Bucket *msg, Peer *p,
		  VisDQMObject &o,
		  const std::string &spec,
		  const std::string &file)
    {
      uint32_t datalen = 0;
      unsigned char *data = 0;

      // Mark the object referenced.
      o.lastreq = Time::current();

      // Validate the image request.
      const char *error = 0;
      VisDQMImgInfo *img = parseImageSpec(o, spec, error);
      if (! img)
      {
	logme()
	  << "ERROR: invalid image specification '"
	  << spec << "' for object '" << o.name
	  << "': " << error << std::endl;
	return;
      }

      // The object is known and the image specification is valid.  Check
      // if we have an already rendered image, and if so, use it as such.
      if (img->version == o.version)
      {
	datalen = img->imgdata.size();
	data = (datalen ? &img->imgdata[0] : 0);
      }

      // We have an object, but no valid image yet.  Render it now.
      else
      {
	doRender(o, *img);
	datalen = img->imgdata.size();
	data = (datalen ? &img->imgdata[0] : 0);
      }

      // Build the response.
      uint32_t words [6];
      uint32_t namelen = o.name.size();
      uint32_t speclen = spec.size();
      uint32_t filelen = file.size();
      words[0] = sizeof(words) + namelen + speclen + filelen + datalen;
      words[1] = DQM_REPLY_IMAGE_FILE;
      words[2] = namelen;
      words[3] = speclen;
      words[4] = filelen;
      words[5] = datalen;

      msg->data.reserve(msg->data.size() + words[0]);
      copydata(msg, &words[0], sizeof(words));
      if (namelen)
	copydata(msg, &o.name[0], namelen);
      if (speclen)
	copydata(msg, &spec[0], speclen);
      if (filelen)
	copydata(msg, &file[0], filelen);
      if (datalen)
	copydata(msg, data, datalen);
    }


  // Parse an arbitrary string, for extracting the file name.
  bool
  parseString(const char *&p, const char *name, size_t len, std::string &value)
    {
      if (! strncmp(p, name, len))
      {
	p += len;
	len = 0;
	while (p[len] && p[len] != ';')
	  ++len;
	value.insert(value.end(), p, p+len);
	p += len;
	return true;
      }
      return false;
    }

  void
  findFileObject(VisDQMObject &o,
		 TObjString &str,
		 const std::string &name,
		 const std::string &filename)
    {
      // If file name is empty, we don't know where the object was.
      if (filename.empty())
        return;

      // Find a TFile for this file name, or open a new one if we
      // don't have this file open yet.  If we have too many files
      // open, close the oldest file.
      FileMap::iterator f = files_.find(filename);
      if (f == files_.end())
      {
	if (files_.size() == maxfiles_)
	{
	  FileMap::iterator oldest = f = files_.begin();
	  while (++f != files_.end())
	    if (f->second.first < oldest->second.first)
	      oldest = f;

	  oldest->second.second->Close();
	  delete oldest->second.second;
	  files_.erase(oldest);
	}

	// Add a new file.
	TFile *file = new TFile(filename.c_str());
	if (file->IsZombie())
	{
	  logme() << "ERROR: cannot open file '" << filename << "'\n";
	  delete file;
	  return;
	}

	logme() << "INFO: opened file '" << filename << "'\n";
	std::pair<Time, TFile *> finfo(Time::current(), file);
	f = files_.insert(FileMap::value_type(filename, finfo)).first;
      }

      // Update the stamp on this file to mark it freshly accessed.
      f->second.first = Time::current();
      TFile *file = f->second.second;

      // FIXME: Quality reports? Object flags?
      // Locate the object in question, patching up all necessary data for it.
      // First try to get the object the cheap way, then try more complexity.
      if (TObject *obj = (TObject *) file->GetObjectChecked(("/DQMData/" + name).c_str(), "TObject"))
      {
	o.version = f->second.first.ns();
        o.object = obj;
        o.reference = (TObject *) file->GetObjectChecked
          (("/DQMData/Reference/" + name).c_str(), "TObject");
      }

      // No luck, it must be a scalar which we have to search by name...
      // We don't bother with reference for scalars.
      size_t slash = name.rfind('/');
      if (slash == name.size()-1)
	return;
      else if (slash == std::string::npos)
	;
      else if (! file->cd(("/DQMData/" + name.substr(0, slash)).c_str()))
	return;

      // Fetch the object.
      const char *namepart = (slash == std::string::npos ? &name[0] : &name[slash+1]);
      o.version = f->second.first.ns();
      std::string scalar;
      scalar.reserve(3 + strlen(namepart));
      scalar += '<';
      scalar += namepart;
      scalar += '>';

      TKey *key;
      TIter next(gDirectory->GetListOfKeys());
      while (! o.object && (key = (TKey *) next()))
      {
	TObject *obj = key->ReadObj();

	// Try a simple, modern scalar.
	if (!strncmp(obj->GetName(), scalar.c_str(), scalar.size()))
	  o.object = obj;

	// If this is a TNamed, fake it into a TObjString.
	else if (TNamed *n = dynamic_cast<TNamed *>(o.object))
	  if (! dynamic_cast<TH1 *>(n))
	  {
	    std::string sv;
	    sv.reserve(6 + strlen(n->GetTitle()) + 2*strlen(n->GetName()));
	    sv += '<'; sv += n->GetName(); sv += '>';
	    sv += n->GetTitle();
	    sv += '<'; sv += '/'; sv += n->GetName(); sv += '>';
	    str = sv.c_str();
	    o.object = &str;
	  }

	else
	  delete obj;
      }
    }
};

class VisDQMArchiveDaemon : public VisDQMRenderDaemon
{
  size_t maxfiles_;

public:
  VisDQMArchiveDaemon(int argc, char **argv)
    : VisDQMRenderDaemon(argc, argv),
      maxfiles_(5)
    {}

protected:
  virtual bool
  parseOption(int &i, int argc, char **argv)
    {
      if (i < argc-1 && !strcmp(argv[i], "--max-files"))
      {
	maxfiles_ = atoi(argv[++i]);
	return true;
      }
      return false;
    }

  virtual bool
  checkOptions(void)
    {
      return maxfiles_ > 0 && maxfiles_ < 100;
    }

  virtual std::string
  showOptions(void)
    {
      return "[--max-files N]";
    }

  virtual void
  runServer(const VisDQMImageServer::ServerInfo &info)
    {
      VisDQMArchiveImageServer server(info, maxfiles_);
      server.run();
    }
};


int main(int argc, char **argv)
{
  VisDQMArchiveDaemon d(argc, argv);
  return d.execute();
}
