import time
from copy import copy,deepcopy
from xml.dom.minidom import parseString
import urllib2

##########################################################
# Start of LHCCooldown example
##########################################################

class LHCSource:
  """
  Our first example. This should be considered a simple, hello-world type exercise.
  The environment we are creating is entirely non-interactive, just serving the LHC
  cooldown state image to the javascript client page.
  """
  
  # Set the plothook global variable - URLs of the form overview/lhc/ will now be directed to this source.
  plothook = 'lhc'
  
  # A dictionary matching up short urls that our javascript will use with the full URL we'll fetch the image from.
  urls = {"lhc.png": "http://hcc.web.cern.ch/hcc/file/field_lhc.png"}
  
  def __init__(self,gui,path,*args):
    """
    Create a new LHCSource. We don't really need to do anything here.
    """
    pass
    
  def plot(self,*path,**attrs):
    """
    Called when a plot request is made of the form overview/lhc/PATH/TO/IMAGE?key=value
    
    In this case we don't really need to worry about arguments since we're only expecting
    one possible value, path=['lhc.png'] and no attributes, but we'll check anyway.
    """
    
    # Check for expected number of arguments
    if len(path)==1:
      # Check it's a short url we know of
      if path[0] in LHCSource.urls:
        mimetype = 'image/png' # We'll just assume they're all png for the moment
        url = LHCSource.urls[path[0]] # Get the full URL from the dictionary
        data = urllib2.urlopen(url).read() # Read the data from this url
        
        # Return the data and mimetype
        return mimetype,data
    
    # Something went wrong, return an error condition
    return None,None 
    
  def prepareSession(self,session):
    """
    Called when a new session hits the server. Nothing to do here.
    """
    pass
  

class LHCWorkspace:
  """
  Workspace for the LHC example. This again is very simple, having only a single view
  and no state updates to worry about. All the necessary methods are implemented, and
  nothing more.
  """
  class CooldownView:
    """
    Inner class representing the cooldown view. Unnecessary when we have only a single view,
    but the convention of having a class per view can be a useful way of handling things. 
    """
    name = 'Cooldown Summary' # The title that will be displayed
    label = 'cooldown' # The name that will be used when view?name= is called
    plugin = 'LHCCooldown' # The name of the plugin that will be loaded for this view

  # Sessiondef contains the defaults that will be copied into each new session, in this
  # case setting the view to 'cooldown'
  sessiondef = {'lhc.view': 'cooldown'}

  def __init__(self,gui,*args):
    """
    Create a new workspace, and store a reference to the server ('gui') object.
    Also create the list of views for the workspace.
    """
    gui._addJSFragment("%s/gui-overview.js" % gui.contentpath)
    gui._addJSFragment("%s/gui-example.js" % gui.contentpath)
    self.gui = gui
    self.name = 'LHC'
    self.views = [LHCWorkspace.CooldownView()]
    # We don't need a list for only a single view, but it's not a bad idea to have one
    # so the following functions will be more generally applicable

  def _state(self,session):
    """
    Internal state method, since we'll be calling this a lot. Creates a JSON representation
    of the session dictionary (the relevant keys, anyway).
    """
    # Get the object representing the current view.
    view = self.views[0]
    for v in self.views:
      if v.label==session['lhc.view']:
        view = v
        break
    
    # Set no-caching headers for the next response.  
    self.gui._noResponseCaching()
    
    # Create the JSON state representation
    # Load the AutoUpdate plugin with 300-sec interval and the current timestamp
    json = "([{kind: 'AutoUpdate', interval: %d, stamp: %d}," % (300, self.gui.stamp)
    # Load the OverviewHeader plugin and populate the lists of workspaces and views using list comprehensions
    json += "{kind: 'OverviewHeader', workspace: 'LHC', workspaces: [%s], view: '%s', views: [%s]}," % \
      (",".join("{title: '%s', label: '%s'}" % (x.name,x.name.lower()) for x in self.gui.workspaces),
      view.name,
      ",".join("{title: '%s', label: '%s'}" % (x.name,x.label) for x in self.views))
    # Load the plugin for the requested view
    # if you want to include any of your own data in the state you need to add code here
    # adding a state method to the view object to provide this line is one way of doing it
    json += "{kind: '%s'}])" % (view.plugin)
      
    # Return the JSON
    return json
    
  def initialiseSession(self,session):
    """
    When the session is created, copy the contents of our sessiondef into it.
    """
    for (key,value) in self.sessiondef.iteritems():
      if key not in session:
        session[key]= copy(value)
        
  def start(self,session,**kwargs):
    """
    When a user first lands on this workspace. Nothing to do here.
    """
    pass
    
  def sessionView(self,session,*args,**kwargs):
    """
    Called when a user changes view with a view?name=newviewname link.
    """
    # Get the view name parameter from the keyword args
    name = kwargs.get('name',None)
    
    #get the object representing the current view, exactly as in _state 
    view = self.views[0]
    for v in self.views:
      if v.label==name:
        view = v
        break
    
    # Set the new view in the session dictionary
    session['lhc.view']=view.label
    
    # Save the session to disk (optional)
    self.gui._saveSession(session)
    
    # Return a string representing the current state, using _state
    return self._state(session)
    
  def sessionState(self,session,*args,**kwargs):
    """
    Called by the frontend to get the current state. Returns _state.
    """
    return self._state(session)
      
  def sessionResetWorkspace(self,session,*args,**kwargs):
    """
    Reset the state back to defaults.
    """
    self.initialiseSession(session)
    self.gui._saveSession(session)
    return self._state(session)
  
##########################################################
# End of LHCCooldown example
##########################################################


##########################################################
# Start of TPGQuery example
##########################################################

class TPGWorkspace:
  """
  Workspace for our second example. This is slightly more complex (handles interaction with javascript),
  but large parts are in common with the previous example and won't be dwelt on.
  """
  class TPGView:
    name = 'TPG Query' # The title that will be displayed
    label = 'query' # The name that will be used when view?name= is called
    plugin = 'TPGQuery' # The name of the plugin that will be loaded for this view

  # Sessiondef contains the defaults that will be copied into each new session, in this
  # case setting the view to 'query'
  # Default line 16662 (56)
  # Default direction r (CERN-Avanchet)
  # Default stop 1 (CERN)
  sessiondef =  {  
          'tpg.view': 'query',
          'tpg.line': '16662',
          'tpg.direction': 'r',
          'tpg.stop': '1'
          }

  def __init__(self,gui,*args):
    gui._addJSFragment("%s/gui-overview.js" % gui.contentpath)
    gui._addJSFragment("%s/gui-example.js" % gui.contentpath)
    self.gui = gui
    self.name = 'TPG'
    self.views = [TPGWorkspace.TPGView()]

  def sessionAlter(self,session,*args,**kwargs):
    """
    Landing function for session-changing calls. When a list value is changed, a call is made to here.
    """
    # Get the field to be altered and the new value from the keywords
    field = kwargs.get('field','')
    value = kwargs.get('value','')
    
    # Validates the field being changed  
    if field in ('line','direction','stop'):
    
      # For change in line, make the current stop and direction null since they're now undefined
      # String 'null' is used as a signalling value for javascript here (as opposed to None)
      # We need to check that the values here are roughly what we expect (to guard against XSS attacks, for instance)
      if field=='line':
        # Value should be an integer ~16000
	if value.isdigit():
          session['tpg.line']=value
          session['tpg.direction']='null'
          session['tpg.stop']='null'
        else:
          return None
      
      # Likewise for direction
      elif field=='direction':
        # Value should be a char 'a' or 'r'
        if value in ('a','r'):
          session['tpg.direction']=value
          session['tpg.stop']='null'
        else:
          return None
      
      # and stop
      elif field=='stop':
        # Value should be a small positive integer
        if value.isdigit():
          session['tpg.stop']=value
        else:
          return None
        
    self.gui._saveSession(session)
    return self._state(session)  

  def _state(self,session):
    """
    Internal state method, since we'll be calling this a lot. Creates a JSON representation
    of the session dictionary (the relevant keys, anyway).
    """
    # Get the object representing the current view, 
    view = self.views[0]
    for v in self.views:
      if v.label==session['lhc.view']:
        view = v
        break
    
    # Set no-caching headers for the next response    
    self.gui._noResponseCaching()
    
    # Create the JSON state representation, first two lines as before
    json = "([{kind: 'AutoUpdate', interval: %d, stamp: %d}," % (300, self.gui.stamp)
    json += "{kind: 'OverviewHeader', workspace: 'LHC', workspaces: [%s], view: '%s', views: [%s]}," % \
      (",".join("{title: '%s', label: '%s'}" % (x.name,x.name.lower()) for x in self.gui.workspaces),
      view.name,
      ",".join("{title: '%s', label: '%s'}" % (x.name,x.label) for x in self.views))
    # Load the plugin for the requested view.
    # This time we need to append a lot of state information to the JSON string
    # for each key tpg.*, append it and a string representation of its value
    # nicely, repr([]) and repr({}) produce JSON-compatible forms
    # (providing the lists/dicts only contain basic types) 
    json += "{kind: '%s', version:0, %s}])" \
      % (view.plugin,
        ",".join("%r:%r" % (key,session[key]) for key in session if key.startswith('tpg.')))
      
    # Return the JSON
    return json
  
  
  ##########################
  # The following functions are identical (or nearly) to examples in LHCWorkspace above
  # and hence undocumented here
  ##########################
  def initialiseSession(self,session):
    for (key,value) in self.sessiondef.iteritems():
      if key not in session:
        session[key]= copy(value)
        
  def start(self,session,**kwargs):
    pass
    
  def sessionView(self,session,*args,**kwargs):
    name = kwargs.get('name',None) 
    view = self.views[0]
    for v in self.views:
      if v.label==name:
        view = v
        break
    session['tpg.view']=view.label
    self.gui._saveSession(session)
    return self._state(session)
    
  def sessionState(self,session,*args,**kwargs):
    return self._state(session)
      
  def sessionResetWorkspace(self,session,*args,**kwargs):
    self.initialiseSession(session)
    self.gui._saveSession(session)
    return self._state(session)

class TPGSource:
  """
  Source for the TPG example. This retrieves data from the backend
  (in this case the TPG AJAX landing page), processes it into an appropriate
  format and serves it.
  """
  plothook = 'tpg'
  
  def __init__(self,gui,path,*args):
    pass
    
  def plot(self,*path,**attrs):
    """
    Return text or image data. In this case we'll only be serving text data. We need to check that
    it is valid, and then retrieve the information from the backend, format it, and serve it.
    
    Expected queries:
    ['linelist']
    ['directionlist'] {'line':'linenum'}
    ['stoplist'] {'line':'linenum', 'direction':'directionchar'}
    ['times'] {'line':'linenum', 'direction':'directionchar', 'stop':'stopnum'}
    """
    
    # Check for right number of arguments
    if len(path)==1:
      path = path[0]
      
      # Check for known arguments
      if path in ('linelist','directionlist','stoplist','times'):
        
        # If we just want the list of lines, return it as a json object
        if path == 'linelist':
          return 'text/plain',"(%r)" % (self._fetchLineList())
          
        # If we want the list of directions, check a line has been supplied
        elif path == 'directionlist':
          if 'line' in attrs:
            return 'text/plain',"(%r)" % (self._fetchDirectionList(attrs['line']))
        
        # For a list of stops, we need a line and direction
        elif path == 'stoplist':
          if 'line' in attrs and 'direction' in attrs:
            return 'text/plain',"(%r)" % (self._fetchStopList(attrs['line'],attrs['direction']))
        
        # and for a list of times we need all three
        elif path == 'times':
          if 'line' in attrs and 'direction' in attrs and 'stop' in attrs:
            return 'text/plain',"(%r)" % (self._fetchResult(attrs['line'],attrs['direction'],attrs['stop']))
      
    # Something went wrong, return nothing (500 Internal Server Error)
    return None,None
  
  ##################
  # How the following methods work is not really important to the example, but
  # it should be fairly simple to see how they fetch XML and then process it with DOM
  ##################
  def _fetchLineList(self):
    return {'16648': '36', '16649': '41', '16660': '54', '16661': '55', '16666': 'd', '16667': 'e', '16664': 'b', '16663': 'a', '16640': '27', '16641': '28', '16642': '29', '16643': '31', '16644': '32', '16645': '33', '16646': '34', '16647': '35', '16672': 's', '16628': '11', '16629': '12', '16626': '9', '16627': '10', '16624': '7', '16625': '8', '16622': '5', '16623': '6', '16620': '2', '16621': '3', '16677': 'z', '16665': 'c', '16676': 'y', '16668': 'f', '16671': 'l', '16670': 'k', '16673': 'v', '16669': 'g', '16675': 'x', '16674': 'w', '16659': '53', '16658': '52', '16657': '51', '16656': '49', '16655': '48', '16654': '46', '16653': '45', '16652': '44', '16651': '43', '16650': '42', '16619': '1', '16639': '23', '16638': '22', '16662': '56', '16635': '19', '16634': '17', '16637': '21', '16636': '20', '16631': '14', '16630': '13', '16633': '16', '16632': '15'} # Static, this could be fetched but storing it here seems fine
  
  def _fetchDirectionList(self,line):
    dir={}
    try:
      data = urllib2.urlopen("http://www.tpg.ch/_services/nextHoraireForm.php?todo=ajaxSearch&lineSelector=%s&t=%s" % (line,time.strftime("%H:%M:%S"))).read()
      dom = parseString(data)
      d_blocks = dom.getElementsByTagName("option")
      
      for b in d_blocks:
        short = b.getAttribute("value").encode('ascii','replace')
        if (len(short.strip())>0):
          desc = b.firstChild.nodeValue.encode('ascii','replace')
          dir[short]=desc
    except URLError:
      return None
    return dir

  def _fetchStopList(self,line,direction):
    stops = {}
    try:
      data = urllib2.urlopen("http://www.tpg.ch/_services/nextHoraireForm.php?todo=ajaxSearch&lineSelector=%s&directionSelector=%s&t=%s" % (line,direction,time.strftime("%H:%M:%S"))).read()
      dom = parseString(data)
      d_blocks = dom.getElementsByTagName("option")

      for b in d_blocks:
        short = b.getAttribute("value").encode('ascii','replace')
        if (len(short.strip())>0):
          desc = b.firstChild.nodeValue.encode('ascii','replace')
          stops[short]=desc
    except URLError:
      return None
    return stops   

  def _fetchResult(self,line,direction,stop):
    times = []
    try:
      data = urllib2.urlopen("http://www.tpg.ch/_services/nextHoraireForm.php?todo=ajaxSearch&lineSelector=%s&directionSelector=%s&arretSelector=%s&t=%s" % (line,direction,stop,time.strftime("%H:%M:%S"))).read()
      data = "<all>%s</all>" % (data)
      dom = parseString(data)

      d_blocks = dom.getElementsByTagName("div")
      
      for b in d_blocks:
        if b.getAttribute("class")=="nextDepartureItem":
          times.append(b.firstChild.nodeValue.encode('ascii','replace'))
    except URLError:
      return None
    return times
     
  def prepareSession(self,session):
    pass
  

##########################################################
# End of TPGQuery example
##########################################################
