from matplotlib.figure import Figure
from matplotlib.transforms import nonsingular,scale_transform
from matplotlib.font_manager import FontProperties
from matplotlib.ticker import Locator, Formatter, ScalarFormatter
from matplotlib.patches import Wedge, Shadow, Rectangle, Polygon, Line2D
from matplotlib.backends.backend_agg import FigureCanvasAgg
from matplotlib.backends.backend_svg import FigureCanvasSVG
from matplotlib.backends.backend_pdf import FigureCanvasPdf
from matplotlib.backends.backend_ps import FigureCanvasPS
from matplotlib.pylab import setp
from matplotlib.numerix import ma
from matplotlib import cm
from cStringIO import StringIO
from time import strftime, gmtime
from tempfile import mkstemp
from GuiUtils import thousands
import math, numpy, os, colorsys, pylab
from PIL import Image

# Utility function to convert text RGB colours to RGB triple in range [0, 1]
def text2rgb(str):
  return [ int(str[i:i+2], 16)/255. for i in xrange(1, len(str), 2) ]

def stitchPlotAndLegend(plot,legend):
  """
  Simple implementation.
  Requires that the 'png' imagetype is used. Extending this for SVG, PDF, PS etc would require presumably
  some digging into the Figure backends for matplotlib, which gets a bit byzantine...
  """
  png_header = '\x89\x50\x4e\x47\x0d\x0a\x1a\x0a'
  if plot.startswith(png_header) and legend.startswith(png_header):
    plotfile = StringIO(plot)
    legendfile = StringIO(legend)
   
    plotimg = Image.open(plotfile)     
    legendimg = Image.open(legendfile)
      
    composite = Image.new(plotimg.mode,(plotimg.size[0],plotimg.size[1]+legendimg.size[1]))
    composite.paste(plotimg,(0,0,plotimg.size[0],plotimg.size[1]))
    composite.paste(legendimg,(0,plotimg.size[1],plotimg.size[0],plotimg.size[1]+legendimg.size[1]))
    f = StringIO()
    composite.save(f,'PNG')
    return f.getvalue()
  else:
    raise NotImplemented
  
# Format axis value as a byte integer number with thousand separators.
class ByteScalarFormatter(Formatter):
  def __call__(self, val, pos=None):
    return thousands("%d" % val)

# Add thousand separators to normal scalar formatter.
class ThousandScalarFormatter(ScalarFormatter):
  def __call__(self, val, pos=None):
    return thousands(ScalarFormatter.__call__(self, val, pos))

# Display a number as a percent.
class PercentFormatter(ScalarFormatter):
  def __call__(self, val, pos=None):
    return "%d%%" % round(val*100)

# For axis range vmin, vmax, finds the largest power of two >= vmax,
# then places ticks linearly between vmin and vmax at N*power-of-two
# positions for integer N such that there are no more than 8 ticks.
# The axis scale range is assumed to be integer and positive.
class ByteScalarLocator(Locator):
  def __call__(self):
    vmin, vmax = self.viewInterval.get_bounds()
    if vmin > vmax:
      vmin, vmax = vmax, vmin

    rmax = (vmax >= 1 and 2**math.ceil(math.log(vmax, 2))) or 0
    rmin = (vmin >= 1 and 2**math.floor(math.log(vmin, 2))) or 0
    step = max(1, int((rmax - rmin)/8))
    rmin += math.ceil((vmin - rmin)/step)
    ticks = numpy.arange(rmin, rmax+0.001, step)
    return ticks

  def autoscale(self):
    self.verify_intervals()
    dmin, dmax = self.dataInterval.get_bounds()
    if dmin > dmax:
      dmin, dmax = dmax, dmin
    if dmin == dmax:
      dmin -= 1
      dmax += 1
    exp, rem = divmod(math.log(dmax - dmin, 2), 1)
    if rem > 0.5:
      exp -= 1
    scale = 2**(-exp)
    dmin = math.floor(scale*dmin)/scale
    dmax = math.ceil(scale*dmax)/scale
    return nonsingular(dmin, dmax)

# Place time series axis tick marks at intervals adapted
# to the time span unit and the length of the axis in time.
class TimeSeriesLocator(Locator):
  def pick(self, series, spec):
    for (interval, roundto) in spec:
      if len(series) / interval < 12:
        break
    base = [i for i in xrange(0, interval) if series[i][0] % roundto == 0][0]
    self.ticks = [t[0] for t in series[base::interval]]

  def __init__(self, span, series):
    if span == 'hour':
      self.pick(series, [(1, 3600), (4, 14400), (8, 14400), (12, 14400), (24, 14400), (48, 86400)])
    elif span == 'day':
      self.pick(series, [(1, 86400), (7, 7*86400), (14, 7*86400), (28, 7*86400), (56, 14*86400)])
    elif span == 'week':
      self.pick(series, [(1, 1), (4, 1), (13, 1), (26, 1), (52, 1)])
    elif span == 'month':
      self.pick(series, [(1, 1), (4, 1), (6, 1), (12, 1), (18, 1), (24, 1), (36, 1), (48, 1)])
    else:
      self.pick(series, [(1, 1), (2, 1), (4, 1), (8, 1), (16, 1), (32, 1), (64, 1)])

  def __call__(self):
    return self.ticks

  def autoscale(self):
    return self.dataInterval.get_bounds()

class TimeSeriesFormatter(Formatter):
  def __init__(self, format):
    self.format = format

  def __call__(self, val, pos=None):
    return strftime(self.format, gmtime(val))

# Chart base class.
class Chart:
  def __init__(self, props, subdefaults):
    self.file = None
    self.filename = None
    self.props = {
      "imageType":	  "png",	# Output image type.
      "dpi":		  72,		# Image DPI resolution.
      "width":		  800,		# Output image width in pixels (vs. DPI).
      "height":		  600,		# Output image height in pixels (vs. DPI).
      "padding":	  50,		# Padding around the figure edge, in pixels.
      "textPadding":	  3,		# Padding around text, in pixels.
      "fontFamily":	  "sans-serif",	# Font family.
      "fontName":	  "Luxi Sans",	# Font name.
      "fontSize":	  12,		# Size of non-title text, in pixels.
      "titleFontName":	  "Bitstream Vera Sans",
      "titleFontSize":	  18,		# Size of title text, in pixels.
      "subtitleFontSize": 14,		# Size of subtitle text, in pixels.
      "axtitleFontSize":  16,		# Size of axis title text, in pixels.
      "xAxisScale":	  "lin",	# X axis scale, lin or log.
      "yAxisScale":	  "lin",	# Y axis scale, lin or log.
      "square":		  False		# Force square layout.
    }
    for (k, v) in subdefaults.iteritems():
      self.props[k] = v
    for (k, v) in props.iteritems():
      self.props[k] = v

    self.font = FontProperties()
    self.font.set_family(self.get('fontFamily'))
    self.font.set_name(self.get('fontName'))
    self.font.set_size(float(self.get('fontSize')))
    self.tfont = FontProperties()
    self.tfont.set_family(self.get('titleFontFamily', self.font.get_family()[-1]))
    self.tfont.set_name(self.get('titleFontName', self.font.get_name()))
    self.tfont.set_size(self.get('titleFontSize', self.font.get_size()))
    self.tfont.set_weight('bold')
    self.sfont = FontProperties()
    self.sfont.set_family(self.get('subtitleFontFamily', self.tfont.get_family()[-1]))
    self.sfont.set_name(self.get('subtitleFontName', self.tfont.get_name()))
    self.sfont.set_size(self.get('subtitleFontSize', self.tfont.get_size()))
    self.afont = FontProperties()
    self.afont.set_family(self.get('axtitleFontFamily', self.tfont.get_family()[-1]))
    self.afont.set_name(self.get('axtitleFontName', self.tfont.get_name()))
    self.afont.set_size(self.get('axtitleFontSize', self.tfont.get_size()))
    self.afont.set_weight('bold')

  def __del__(self):
    if self.file != None:
      os.close(self.file)

  def get(self, key, default=None):
    return getattr(self, key, self.props.get(key, default))

  def draw(self):
    (fig, canvas, w, h) = self.canvas()
    fig.text(.5, .5, "No data.", horizontalalignment='center', fontproperties=self.font)
    return self.save(fig, canvas)

  def legend(self):
    (fig, canvas, w, h) = self.canvas()
    fig.text(.5, .5, "No legend.", horizontalalignment='center', fontproperties=self.font)
    return self.save(fig, canvas)

  def details(self):
    return {}

  def canvas(self):
    type = self.get("imageType", "png")
    fig = Figure()
    if type == "png":
      canvas = FigureCanvasAgg(fig)
      (self.file, self.filename) = mkstemp(".%s" % type)
    elif type == "svg":
      canvas = FigureCanvasSVG(fig)
      (self.file, self.filename) = mkstemp(".%s" % type)
    elif type == "pdf":
      canvas = FigureCanvasPdf(fig)
      (self.file, self.filename) = mkstemp(".%s" % type)
    elif type == "ps" or type == "eps":
      canvas = FigureCanvasPS(fig)
      (self.file, self.filename) = mkstemp(".%s" % type)
    else:
      raise "Invalid render target requested"

    # Set basic figure parameters
    dpi = float(self.get('dpi'))
    (w, h) = (float(self.get('width')), float(self.get('height')))
    (win, hin) = (w/dpi, h/dpi)
    fig.set_size_inches(win, hin)
    fig.set_dpi(dpi)
    fig.set_facecolor('white')
    return (fig, canvas, w, h)

  def save(self, fig, canvas):
    canvas.draw()
    if not self.filename:
      canvas.print_figure(self.file, dpi=float(self.get('dpi')))
      return self.file
    else:
      canvas.print_figure(self.filename, dpi=float(self.get('dpi')))
      f = file(self.filename, "r") # PS backend writes over the file
      os.remove(self.filename)     #    return os.fdopen(self.file)
      os.close(self.file)
      self.file = None
      self.filename = None
      return f

  def prepare(self):
    # Create canvas and determine figure parameters
    (fig, canvas, w, h) = self.canvas()
    dpif = float(self.get('dpi')) / 72
    padding = float(self.get('padding')) * dpif / h
    textPadding = float(self.get('textPadding')) * dpif
    titleFontSize = float(self.get('titleFontSize'))
    axFontSize = float(self.get('axtitleFontSize'))
    subFontSize = float(self.get('subtitleFontSize'))
    title = self.get("title", "").split("\n")
    hsub = (len(title)-1) * (subFontSize * dpif + 4) + textPadding
    htitle = hsub + titleFontSize * dpif + textPadding * 2

    # Configure axes
    if self.get('square'):
      minsize = 1 - 2*padding
      axrect = (.5 - minsize/2 * h/w, padding, h/w * minsize, minsize)
    else:
      axrect = (padding, padding, 1 - 1.25*padding, 1 - htitle/h - padding)
    ax = fig.add_axes(axrect)
    frame = ax.get_frame() 
    frame.set_fill(False)

    xlog = (str(self.get('xAxisScale', "lin")) == 'log')
    ylog = (str(self.get('yAxisScale', "lin")) == 'log')
    if xlog:
      ax.semilogx()
    if ylog:
      ax.semilogy()

    setp(ax.get_xticklabels(), fontproperties=self.font)
    setp(ax.get_yticklabels(), fontproperties=self.font)
    setp(ax.get_xticklines(), markeredgewidth=2.0, zorder=4.0)
    setp(ax.get_yticklines(), markeredgewidth=2.0)
    ax.grid(True, alpha=0.25, color='#000000', linewidth=0.1)

    # Set titles
    ax.title = ax.text(.5, 1+(hsub+textPadding)/(axrect[-1]*h), title[0],
                      verticalalignment='bottom',
                      horizontalalignment='center',
                      transform=ax.transAxes,
                      clip_box=None,
                      fontproperties=self.tfont)
    ax._set_artist_props(ax.title)
    if len(title) > 1:
      ax.subtitle = ax.text(.5, 1+textPadding/(axrect[-1]*h), "\n".join(title[1:]),
                           verticalalignment='bottom',
                           horizontalalignment='center',
                           transform=ax.transAxes,
                           clip_box=None,
                           fontproperties=self.sfont)

    ax.set_xlabel(self.get("xAxisTitle", ""), fontproperties=self.afont)
    ax.set_ylabel(self.get("yAxisTitle", ""), fontproperties=self.afont)
    return (fig, canvas, ax)
    
  def map(self):
    return ''

  def make(self):
    img = self.draw().read()
    legend = self.legend().read()
    map = self.map()
    return img,legend,map

class PieChart(Chart):
  defaults = {}
  def __init__(self,data,legend,props):
    Chart.__init__(self,props,self.defaults)
    self._data = data
    self._legend = legend
    self._map = {}
    self.title = "%s\nfrom %s to %s UTC" % (self.get('title', ''),
                  strftime("%A %Y-%m-%d %H:%M", gmtime(props['starttime'])),
                  strftime("%A %Y-%m-%d %H:%M", gmtime(props['endtime'])))
        
  def draw(self):
    if len(self._data) == 0:
      return Chart.draw(self)
    fig,canvas,ax = self.prepare()
        
    getcol = lambda x: [item['attrs']['facecolor'] for item in self._legend if item['name']==x][0]
        #data = [(d,self._data[d],getcol(d)) for d in self._data]
        
    labels = []
    values = []
    colours = []
        
    for d in self._data:
      labels.append(d)
      values.append(self._data[d])
      colours.append(getcol(d))
    
    slices, texts = ax.pie(values,None,None,colours)
        
    # The list of vertices from a Wedge contains vertices at 0.01 radian intervals
    # reduce the vertex list with every nth point from the curved segment
    def vert_reduce(x):
      interval = 40
      verts = [x[0]]
      verts.extend(x[1::interval])
      if not len(x) % interval == 2:
        verts.append(x[-1])
    
      # Flatten the list from ((x,y),(x,y)... to (x,y,x,y...
      flat = []
      for v in verts:
        flat.extend((v[0],float(self.get('height')) - v[1]))
      return flat
        
    for l,s in zip(labels,slices):
      verts = s.get_transform().seq_xy_tups(s.get_verts())
      self._map[l]=vert_reduce(verts)
        
    return self.save(fig,canvas)
        
    
  def prepare(self):
    (fig, canvas, w, h) = self.canvas()
    dpif = float(self.get('dpi')) / 72
    padding = float(self.get('padding')) * dpif / h
    textPadding = float(self.get('textPadding')) * dpif
    titleFontSize = float(self.get('titleFontSize'))
    axFontSize = float(self.get('axtitleFontSize'))
    subFontSize = float(self.get('subtitleFontSize'))
    title = self.get("title", "").split("\n")
    hsub = (len(title)-1) * (subFontSize * dpif + 4) + textPadding
    htitle = hsub + titleFontSize * dpif + textPadding * 2

    # Configure axes
    if self.get('square'):
      minsize = 1 - 2*padding
      axrect = (.5 - minsize/2 * h/w, padding, h/w * minsize, minsize)
    else:
      axrect = (padding, padding, 1 - 1.25*padding, 1 - htitle/h - padding)
    ax = fig.add_axes(axrect)
    frame = ax.get_frame() 
    frame.set_fill(False)
  
    
    ax.title = ax.text(.5, 1+(hsub+textPadding)/(axrect[-1]*h), title[0],
                       verticalalignment='bottom',
                       horizontalalignment='center',
                       transform=ax.transAxes,
                       clip_box=None,
                       fontproperties=self.tfont)
    ax._set_artist_props(ax.title)
    if len(title) > 1:
      ax.subtitle = ax.text(.5, 1+textPadding/(axrect[-1]*h), "\n".join(title[1:]),
                            verticalalignment='bottom',
                            horizontalalignment='center',
                            transform=ax.transAxes,
                            clip_box=None,
                            fontproperties=self.sfont)

    return (fig, canvas, ax)
  
  def map(self):
    result = ""
    
    if len(self._map) == 0 and len(self._data) > 0:
      self.draw()
    
    polyStr = lambda x: ",".join([str(int(p)) for p in x])
    mapFormat = lambda fstr,name,value: fstr.replace('$n',str(name)).replace('$v',str(value))
    
    for name in self._map:
      value = self._data[name]
      attribs = []
      for aname in [n for n in self.props if n.startswith('map_')]:
        aval = self.get(aname)
        if aval:
          attribs.append('%s="%s"' % (aname[4:],mapFormat(aval,name,value)))
      result += """<area %s shape="polygon" coords="%s">\n""" % (' '.join(attribs),polyStr(self._map[name]))
    
    return result
    
  
  def legend(self):
    i = 1
    namelen = 0
    legend = []
    for item in self._legend:
      name = "%s [%.2e]" % (item['name'],self._data[item['name']])
      legend.append((i, name, item['attrs']))
      namelen = max(namelen, len(name))
      i += 1

    dpif = float(self.get('dpi')) / 72
    wcol = 10 + (2 + 0.6*namelen)*self.font.get_size()*dpif
    ncol = max(1, math.floor(float(self.get('width')) / wcol))
    wcol = float(self.get('width')) / ncol
    nrow = int((i + ncol - 1) / ncol)
    self.height = (nrow + 1) * self.font.get_size() * 1.4 * dpif
    (fig, canvas, w, h) = self.canvas()
    step = self.font.get_size() * 1.4 * dpif / h
    x = 0
    y = 0.5 * self.font.get_size() * dpif / h
    xbox = 10./w
    wbox = self.font.get_size() * dpif / w * 1.25
    hbox = self.font.get_size() * dpif / h * 0.75
    xtext = xbox + wbox*1.5
    row = 0
    for item in legend:
      fig.text(x + xtext, 1 - y, item[1],
           horizontalalignment='left',
           verticalalignment='top',
           fontproperties=self.font)
      p = Rectangle((x + xbox, 1-y-hbox), wbox, hbox, linewidth=0.5, fill=True)
      p.update(item[2])
      fig._set_artist_props(p)
      fig.patches.append(p)
      y += step
      row += 1
      if row == nrow:
        x += wcol/w
        y = 0.5 * self.font.get_size() * dpif / h
        row = 0
    return self.save(fig, canvas)

class CumulativeTimeSeriesChart(Chart):
  defaults={}
  def __init__(self,span,series,data,legend,props):
    Chart.__init__(self,props,self.defaults)
    self._span = span
    self._series = series
    self._data = data
    self._legend = legend
    self._map = {}
    self._integral = {}
    if span == 'hour':
      self._axisFormat = "%H:%M\n%a %d"
      spanFormat = "%A %Y-%m-%d %H:%M"
    elif span == 'day':
      self._axisFormat = "%a %d\n%b %Y"
      spanFormat = "%A %Y-%m-%d"
    elif span == 'week':
      self._axisFormat = "Week %V\n%d %b, %Y"
      spanFormat = "%Y-%V"
    elif span == 'month':
      self._axisFormat = "%B\n%Y"
      spanFormat = "%Y-%m"
    elif span == 'year':
      self._axisFormat = "%Y"
      spanFormat = "%Y"
    else:
      raise "Bad time series span"

    self.title = "%s\n%d %ss from %s to %s UTC" % \
      (self.get('title', ''), len(series), span,
       strftime(spanFormat, gmtime(series[0][0])),
       strftime(spanFormat, gmtime(series[-1][0])))
      
  def prepare(self):
    fig,canvas,ax = Chart.prepare(self)
    if len(self._series) > 0:
      ax.set_xlim(xmin = self._series[0][0], xmax = self._series[-1][1])
      ax.xaxis.set_major_locator(TimeSeriesLocator(self._span, self._series))
      ax.xaxis.set_major_formatter(TimeSeriesFormatter(self._axisFormat))
      ax.xaxis.set_clip_on(False)
      ax.yaxis.set_major_formatter(ThousandScalarFormatter())
    return (fig, canvas, ax)
  
  def draw(self):
    if len(self._series) == 0:
      return Chart.draw(self)

    keys = []
    color = {}
    for item in self._legend:
      keys.append(item['key'])
      color[item['key']] = item['attrs']['facecolor']
    keys = sorted(keys)
    

    xcoords = numpy.zeros(len(self._series)+1)
    xcoords[0] = self._series[0][0]
    for i,s in enumerate(self._series):
      xcoords[i+1] = s[1]

    (fig, canvas, ax) = self.prepare()
    surfaces = []
    surfaces.append(numpy.zeros(len(self._series)+1))
    
    for k in keys:
      self._integral[k] = 0.
      cur = numpy.zeros(len(self._series)+1)
      total = 0
      for i,d in enumerate(self._data):
        if k in d:
          value = d[k]
          self._integral[k] += value
          total += value
        cur[i+1] = surfaces[len(surfaces)-1][i+1] + total
      surfaces.append(cur)
    
    xcoords2 = numpy.zeros(2*(len(self._series)+1))
    for i,x in enumerate(xcoords):
      xcoords2[i] = x
    for i,x in enumerate(xcoords[::-1]):
      xcoords2[i+len(xcoords)] = x
    
    map_patches = {}
    for i,k in enumerate(keys):
      surface2 = numpy.zeros(len(xcoords2))
      for j,s in enumerate(surfaces[i]):
        surface2[j] = s
      
      for p,q in enumerate(surfaces[i+1][::-1]):
        surface2[p+len(xcoords)] = q
      
      # This function (fill_between) is unavailable at this version of matplotlib, so we need to make complete polygons
      # patch = ax.fill_between(xcoords,surfaces[i],surfaces[i+1],True,facecolor=color[k],closed=True)
      patches = ax.fill(xcoords2,surface2,facecolor=color[k])
      map_patches[k] = patches[0]
    
    for k in map_patches:
      p = map_patches[k]
      maparea = []
      
      def reduce_verts(verts):
        if len(verts) >= 4:
          if not (verts[0][0] == verts[-1][0] and verts[0][1] == verts[-1][1]):
            return verts
          if not (verts[1][0] == verts[-2][0] and verts[1][1] == verts[-2][1]):
            return verts
          # First two points are co-located - ie zero area, truncate them
          verts=verts[1:-1]
          return reduce_verts(verts)
        # If there are less than 4, then there must be only two points (at the same x) left - pointless
        return None
      # There is only ever a single patch here
      verts = reduce_verts(p.get_transform().seq_xy_tups(p.get_verts()))
      if verts:
        for v in verts:
          maparea.extend((v[0],float(self.get('height')) - v[1]))
        self._map[k] = maparea
     
    return self.save(fig,canvas)
    
    
    
  
  def legend(self):
    seen = {}
    for d in self._data:
      for k in d:
        seen[k] = 1

    i = 1
    namelen = 0
    legend = []
    for item in self._legend:
      if not item['key'] or item['key'] not in seen: continue
      name = "%s [%.2e]" % (item['name'],self._integral[item['name']])
      legend.append((i, name, item['attrs']))
      namelen = max(namelen, len(name))
      i += 1

    dpif = float(self.get('dpi')) / 72
    wcol = 10 + (2 + 0.6*namelen)*self.font.get_size()*dpif
    ncol = max(1, math.floor(float(self.get('width')) / wcol))
    wcol = float(self.get('width')) / ncol
    nrow = int((i + ncol - 1) / ncol)
    self.height = (nrow + 1) * self.font.get_size() * 1.4 * dpif
    (fig, canvas, w, h) = self.canvas()
    step = self.font.get_size() * 1.4 * dpif / h
    x = 0
    y = 0.5 * self.font.get_size() * dpif / h
    xbox = 10./w
    wbox = self.font.get_size() * dpif / w * 1.25
    hbox = self.font.get_size() * dpif / h * 0.75
    xtext = xbox + wbox*1.5
    row = 0
    for item in legend:
      fig.text(x + xtext, 1 - y, item[1],
         horizontalalignment='left',
         verticalalignment='top',
         fontproperties=self.font)
      p = Rectangle((x + xbox, 1-y-hbox), wbox, hbox, linewidth=0.5, fill=True)
      p.update(item[2])
      fig._set_artist_props(p)
      fig.patches.append(p)
      y += step
      row += 1
      if row == nrow:
        x += wcol/w
        y = 0.5 * self.font.get_size() * dpif / h
        row = 0
    return self.save(fig, canvas)
  
  def map(self):
    result = ""
  
    if len(self._map) == 0 and len(self._data) > 0:
      self.draw()
  
    polyStr = lambda x: ",".join([str(int(p)) for p in x])
    mapFormat = lambda fstr,name,value: fstr.replace('$n',str(name)).replace('$v',str(value))
  
    for name in self._map:
      value = self._integral[name]
      attribs = []
      for aname in [n for n in self.props if n.startswith('map_')]:
        aval = self.get(aname)
        if aval:
          attribs.append('%s="%s"' % (aname[4:],mapFormat(aval,name,value)))
      result += """<area %s shape="polygon" coords="%s">\n""" % (' '.join(attribs),polyStr(self._map[name]))
      
    return result
  

# Time series stacked bar chart.
class TimeSeriesBarChart(Chart):
  defaults = {}
  def __init__(self, span, series, data, legend, props):
    Chart.__init__(self, props, self.defaults)
    self._span = span
    self._series = series
    self._data = data
    self._legend = legend
    self._map = []
    self._integral = {}

    if span == 'hour':
      self._axisFormat = "%H:%M\n%a %d"
      spanFormat = "%A %Y-%m-%d %H:%M"
    elif span == 'day':
      self._axisFormat = "%a %d\n%b %Y"
      spanFormat = "%A %Y-%m-%d"
    elif span == 'week':
      self._axisFormat = "Week %V\n%d %b, %Y"
      spanFormat = "%Y-%V"
    elif span == 'month':
      self._axisFormat = "%B\n%Y"
      spanFormat = "%Y-%m"
    elif span == 'year':
      self._axisFormat = "%Y"
      spanFormat = "%Y"
    else:
      raise "Bad time series span"

    self.title = "%s\n%d %ss from %s to %s UTC" % \
      (self.get('title', ''), len(series), span,
       strftime(spanFormat, gmtime(series[0][0])),
       strftime(spanFormat, gmtime(series[-1][0])))

  def prepare(self):
    (fig, canvas, ax) = Chart.prepare(self)
    if len(self._series) > 0:
      ax.set_xlim(xmin = self._series[0][0], xmax = self._series[-1][1])
      ax.xaxis.set_major_locator(TimeSeriesLocator(self._span, self._series))
      ax.xaxis.set_major_formatter(TimeSeriesFormatter(self._axisFormat))
      ax.xaxis.set_clip_on(False)
      ax.yaxis.set_major_formatter(ThousandScalarFormatter())
    return (fig, canvas, ax)

  def draw(self):
    if len(self._series) == 0:
      return Chart.draw(self)

    (fig, canvas, ax) = self.prepare()
    spatch=[]
    for item in self._legend:
      self._integral[item['key']]=0
    
    for i in xrange(0, len(self._series)):
      smap = {}
      
      bin = self._series[i]
      data = self._data[i]
      if len(data) == 0:
        data = { None: 0 }
      l = numpy.zeros(len(data))
      b = numpy.zeros(len(data))
      h = numpy.zeros(len(data))
      a = [None for x in data]
      n = [None for x in data]
      j = 0
      for item in self._legend:
        if item['key'] in data:
          l[j] = bin[0]
          b[j] = (j > 0 and b[j-1] + h[j-1]) or 0
          h[j] = data[item['key']]
          self._integral[item['key']]+=data[item['key']]
          a[j] = item['attrs']
          n[j] = item['key']
          j += 1
      p = ax.bar(l, h, bottom=b, width=bin[1]-bin[0], linewidth=.5)
      for i in xrange(0, len(p)):
        p[i].update(a[i] or {})
        smap[n[i]]=p[i]
      spatch.append(smap)
    for s in spatch:
      sresult={}
      for key in s:
        patch = s[key]
        verts = patch.get_transform().seq_xy_tups(patch.get_verts())
        maparea=[]
        for v in verts[::2]:
          maparea.extend((v[0],float(self.get('height'))-v[1]))
        sresult[key]=maparea
      self._map.append(sresult)
    return self.save(fig, canvas)

  def legend(self):
    seen = {}
    for d in self._data:
      for k in d:
        seen[k] = 1

    i = 1
    namelen = 0
    legend = []
    for item in self._legend:
      if not item['key'] or item['key'] not in seen: continue
      name = "%s [%.2e]" % (item['name'],self._integral[item['name']])
      legend.append((i, name, item['attrs']))
      namelen = max(namelen, len(name))
      i += 1

    dpif = float(self.get('dpi')) / 72
    wcol = 10 + (2 + 0.6*namelen)*self.font.get_size()*dpif
    ncol = max(1, math.floor(float(self.get('width')) / wcol))
    wcol = float(self.get('width')) / ncol
    nrow = int((i + ncol - 1) / ncol)
    self.height = (nrow + 1) * self.font.get_size() * 1.4 * dpif
    (fig, canvas, w, h) = self.canvas()
    step = self.font.get_size() * 1.4 * dpif / h
    x = 0
    y = 0.5 * self.font.get_size() * dpif / h
    xbox = 10./w
    wbox = self.font.get_size() * dpif / w * 1.25
    hbox = self.font.get_size() * dpif / h * 0.75
    xtext = xbox + wbox*1.5
    row = 0
    for item in legend:
      fig.text(x + xtext, 1 - y, item[1],
              horizontalalignment='left',
              verticalalignment='top',
              fontproperties=self.font)
      p = Rectangle((x + xbox, 1-y-hbox), wbox, hbox, linewidth=0.5, fill=True)
      p.update(item[2])
      fig._set_artist_props(p)
      fig.patches.append(p)
      y += step
      row += 1
      if row == nrow:
        x += wcol/w
        y = 0.5 * self.font.get_size() * dpif / h
        row = 0
    return self.save(fig, canvas)

  def map(self):
    result = ""
    polyStr = lambda x: ",".join([str(int(p)) for p in x])
    for i,data in enumerate(self._data):
      starttime = self._series[i][0]
      endtime = self._series[i][1]
      
      mapFormat = lambda fstr,name,value: fstr.replace('$n',str(name)).replace('$v',str(value)).replace('$b',strftime("%H:%M", gmtime(starttime))).replace('$e',strftime("%H:%M", gmtime(endtime)))
      
      for name in data:
        value = data[name]
        if value>0:
          attribs = []
          for aname in [n for n in self.props if n.startswith('map_')]:
            aval = self.get(aname)
            if aval:
              attribs.append('%s="%s"' % (aname[4:],mapFormat(aval,name,value)))
          result += """<area %s shape="rect" coords="%s">\n""" % (' '.join(attribs),polyStr(self._map[i][name]))
          
    return result
    

class TimeSeriesByteBarChart(TimeSeriesBarChart):
  def prepare(self):
    (fig, canvas, ax) = TimeSeriesBarChart.prepare(self)
    ax.yaxis.set_major_locator(ByteScalarLocator())
    ax.yaxis.set_major_formatter(ByteScalarFormatter())
    return (fig, canvas, ax)

class RateQualityScatterChart(Chart):
  defaults = {}
  def __init__(self, span, series, data, legend, props):
    Chart.__init__(self, props, self.defaults)
    self._span = span
    self._series = series
    self._data = data
    self._legend = legend

    if span == 'hour':
      spanFormat = "%A %Y-%m-%d %H:%M"
    elif span == 'day':
      spanFormat = "%A %Y-%m-%d"
    elif span == 'week':
      spanFormat = "%Y-%V"
    elif span == 'month':
      spanFormat = "%Y-%m"
    elif span == 'year':
      spanFormat = "%Y"
    else:
      raise "Bad time series span"

    self.title = "%s\n%d %ss from %s to %s UTC" % \
      (self.get('title', ''), len(series), span,
       strftime(spanFormat, gmtime(series[0][0])),
       strftime(spanFormat, gmtime(series[-1][0])))

  def prepare(self):
    (fig, canvas, ax) = Chart.prepare(self)
    ax.xaxis.set_major_formatter(PercentFormatter())
    ax.yaxis.set_major_formatter(ThousandScalarFormatter())
    return (fig, canvas, ax)

  def draw(self):
    if len(self._series) == 0 or len(self._data) == 0:
      return Chart.draw(self)

    (fig, canvas, ax) = self.prepare()
    i = ymax = 0
    x = numpy.zeros(len(self._data))
    y = numpy.zeros(len(self._data))
    s = numpy.zeros(len(self._data))
    c = ['#000000' for item in self._data]
    t = self.get('letters', False)
    for item in self._legend:
      if item['key'] not in self._data:
        continue
      (rate, qval, n) = self._data[item['key']]
      ymax = max(ymax, rate)
      x[i] = qval
      y[i] = rate
      s[i] = 5 * (1 + math.log(n, 10))
      c[i] = item['attrs']['facecolor']
      i += 1
      if t:
        bright = colorsys.rgb_to_yiq(*text2rgb(c[i-1]))[0] > 0.6
        ax.text(qval, rate, str(i),
               verticalalignment='center',
               horizontalalignment='center',
               clip_box = None,
               color=(bright and '#000000') or '#ffffff',
               fontproperties=self.font)
    if self.get('bubbles', True):
      ax.scatter(x, y, s * s, facecolor=c)

    if self.get('bubbles', True):
      ymin, ymax = ax.get_ylim()
    else:
      ymax = ymax * 1.1
    ax.set_xlim(xmin=-0.01, xmax=1.01)
    ax.set_ylim(ymin=-ymax/100, ymax=ymax)
    return self.save(fig, canvas)

  def legend(self):
    i = 1
    namelen = 0
    legend = []
    seen = dict((k, 1) for k in self._data)
    for item in self._legend:
      if not item['key'] or item['key'] not in seen: continue
      legend.append((i, item['name'], item['attrs']))
      namelen = max(namelen, len(item['name']))
      i += 1
   
    dpif = float(self.get('dpi')) / 72
    wcol = 10 + (5 + 0.6*namelen)*self.font.get_size() * dpif
    ncol = max(1, math.floor(float(self.get('width')) / wcol))
    wcol = float(self.get('width')) / ncol
    nrow = int((i + ncol - 1) / ncol)
    self.height = (nrow + 1) * self.font.get_size() * 1.4 * dpif
    (fig, canvas, w, h) = self.canvas()
    x = 0
    y = self.font.get_size() * 0.5 * dpif / h
    step = self.font.get_size() * 1.4 * dpif / h
    xnr = 10./w
    xbox = xnr + math.ceil(math.log(i, 10)) * self.font.get_size() * dpif / w
    wbox = self.font.get_size() * dpif / w * 1.25
    hbox = self.font.get_size() * dpif / h * 0.75
    xtext = xbox + wbox*1.5
    row = 0
    for item in legend:
      fig.text(x + xnr, 1 - y, str(item[0]) + ".",
              horizontalalignment='left',
              verticalalignment='top',
              fontproperties=self.font)
      fig.text(x + xtext, 1 - y, item[1],
              horizontalalignment='left',
              verticalalignment='top',
              fontproperties=self.font)
      p = Rectangle((x + xbox, 1-y-hbox), wbox, hbox, linewidth=0.5, fill=True)
      p.update(item[2])
      fig._set_artist_props(p)
      fig.patches.append(p)
      y += step
      row += 1
      if row == nrow:
        x += wcol/w
        y = 0.5 * self.font.get_size() * dpif / h
        row = 0
    return self.save(fig, canvas)

class LatencyDensityMap(Chart):
  defaults = {}
  def __init__(self, span, series, data, legend, props):
    Chart.__init__(self, props, self.defaults)
    self._span = span
    self._series = series
    self._data = data
    self._legend = legend

    if span == 'hour':
      spanFormat = "%A %Y-%m-%d %H:%M"
    elif span == 'day':
      spanFormat = "%A %Y-%m-%d"
    elif span == 'week':
      spanFormat = "%Y-%V"
    elif span == 'month':
      spanFormat = "%Y-%m"
    elif span == 'year':
      spanFormat = "%Y"
    else:
      raise "Bad time series span"

    self.title = "%s\n%d %ss from %s to %s UTC" % \
      (self.get('title', ''), len(series), span,
       strftime(spanFormat, gmtime(series[0][0])),
       strftime(spanFormat, gmtime(series[-1][0])))

    self.keys = []
    for item in self._legend:
      if item['key'] in self._data:
        self.keys.append(item)

  def draw(self):
    if len(self._series) == 0 \
       or len(self._data) == 0 \
       or len(self.keys) == 0:
      return Chart.draw(self)

    (fig, canvas, ax) = self.prepare()
    w, h = float(self.get('width')), float(self.get('height'))
    dpif = float(self.get('dpi')) / 72
    axrect = ax.get_position()
    nkeys = len(self.keys)
    cmap = cm.Paired

    tickfont = self.font
    wlabels = nkeys*tickfont.get_size()*dpif
    if wlabels / w > (axrect[2]-axrect[0]) / 1.2:
      tickfont = tickfont.copy()
      tickfont.set_size(max(5, w * (axrect[2]-axrect[0]) / nkeys / 1.2))

    maxname = max(len(i['name']) for i in self.keys)
    maxname = 0.6*(1 + maxname)*tickfont.get_size()*dpif/h
    cbartitle = self.afont.get_size()*dpif/h;
    ax.set_position((axrect[0], maxname,
                    axrect[2]-cbartitle,
                    axrect[3]+axrect[1]-maxname))

    valuef = self.get('yAxisValue')
    nbins = self.get('yAxisBins')
    xmax = max(nkeys, 16)
    x, y = numpy.meshgrid(numpy.arange(xmax+1), [valuef(bin) for bin in xrange(0, nbins+1)])
    c = numpy.zeros((nbins+1, xmax))
    for i in xrange(0, nkeys):
      item = self.keys[i]
      d = self._data[item['key']]
      hmin = hmax = None
      for hbin, n in d.iteritems():
        c[hbin, i] = n
        if hmin == None:
          hmin = valuef(hbin)
          hmax = valuef(hbin+1)
        else:
          hmin = min(hmin, valuef(hbin))
          hmax = max(hmax, valuef(hbin+1))
      if hmin != None:
        ax.axhline(hmin, float(i)/xmax, float(i+1)/xmax, color="#000000", linewidth=.95)
        ax.axhline(hmax, float(i)/xmax, float(i+1)/xmax, color="#000000", linewidth=.95)

    map = ax.pcolor(x, y, ma.masked_where(c == 0, c), shading='flat', cmap=cmap)
    cb = fig.colorbar(map, aspect=50, fraction=0.025, pad=0.01)
    cb.ax.set_ylabel(self.get("yScaleAxisTitle", ""), fontproperties=self.afont)
    setp(cb.ax.get_yticklabels(), fontproperties=self.font)

    setp(ax.get_xticklines(), markeredgewidth=0, zorder=4.0)
    ax.set_xticks(numpy.arange(0, xmax, 1) + 0.5)
    ax.set_xticklabels([x['name'] for x in self.keys],
                      rotation='vertical',
                      horizontalalignment='center',
                      fontproperties=tickfont)
    ax.set_xlim(xmin=0, xmax=xmax)
    ax.set_ylim(ymin=valuef(0), ymax=valuef(nbins+1))
    return self.save(fig, canvas)

  def legend(self):
    dpif = float(self.get('dpi')) / 72
    self.height = 1 * self.font.get_size() * 1.4 * dpif
    (fig, canvas, w, h) = self.canvas()
    fig.text(.5, .5, "No legend",
            horizontalalignment='center',
            verticalalignment='center')
    return self.save(fig, canvas)
