#include "VisMonitoring/DQMServer/interface/VisDQMRenderDaemon.h"
#include "VisMonitoring/DQMServer/interface/VisDQMImageServer.h"
#include "classlib/utils/DebugAids.h"
#include "classlib/utils/TimeInfo.h"
#include "classlib/utils/Time.h"
#include "classlib/utils/Signal.h"
#include "classlib/utils/StringOps.h"
#include "classlib/utils/StringList.h"
#include "classlib/iobase/Filename.h"
#include "TROOT.h"
#include <map>
#include <fstream>
#include <iostream>
#include <cerrno>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/mman.h>

using namespace lat;

VisDQMRenderDaemon::VisDQMRenderDaemon(int argc, char **argv)
  : argv_(argv), argc_(argc)
{
  DebugAids::failHook(&onAssertFail);
  Signal::handleFatal(argv[0], IOFD_INVALID, 0, 0, FATAL_OPTS);
}

VisDQMRenderDaemon::~VisDQMRenderDaemon(void)
{
}

int
VisDQMRenderDaemon::execute(void)
{
  // Check and process arguments.
  VisDQMImageServer::ServerInfo params;
  params.argc = argc_;
  params.argv = argv_;
  params.port = 9091;
  params.debug = false;
  params.crashobj = 0;
  bool daemonise = true;
  for (int i = 1; i < argc_; ++i)
    if (i < argc_-1 && ! strcmp(argv_[i], "--state-directory"))
      params.statedir = argv_[++i];
    else if (i < argc_-1 && ! strcmp(argv_[i], "--listen"))
      params.port = atoi(argv_[++i]);
    else if (i < argc_-1 && ! strcmp(argv_[i], "--black-list"))
      params.blacklistFile = argv_[++i];
    else if (! strcmp(argv_[i], "--no-debug"))
      params.debug = false;
    else if (! strcmp(argv_[i], "--debug"))
      params.debug = true;
    else if (! strcmp(argv_[i], "--daemon"))
      daemonise = true;
    else if (! strcmp(argv_[i], "--no-daemon"))
      daemonise = false;
    else if (! strncmp(argv_[i], "--", 2)
	     && parseOption(i, argc_, argv_))
      ;
    else
      break;

  if (! params.statedir || ! *params.statedir || ! params.port)
  {
    std::cerr << "Usage: " << argv_[0]
	      << " --state-directory STATE-DIRECTORY"
	      << " [--listen PORT]"
	      << " [--[no-]debug] [--[no-]daemon]"
	      << showOptions()
	      << " [render-plugin-options]\n";
    return 1;
  }

  if (! params.statedir.isDirectory ())
  {
    std::cerr << params.statedir << ": not a directory\n";
    return 1;
  }

  if (! params.blacklistFile || ! *params.blacklistFile)
    params.blacklistFile = Filename(params.statedir, "blacklist.txt");

  // Allocate a shared memory segment shared between the
  // the parent and the child processes.  The child will
  // store in this segment the name of the object it was
  // last rendering; if the child crashes the parent adds
  // the name to the list of black-listed objects.
  params.crashobj = (char *) mmap(0, VisDQMImageServer::CRASH_NAME_LEN,
				  PROT_READ | PROT_WRITE,
				  MAP_SHARED | MAP_ANONYMOUS, -1, 0);
  if (params.crashobj == MAP_FAILED)
  {
    logme()
      << "ERROR: failed to allocate shared memory for black listing: "
      << strerror(errno) << " (error code " << errno << ")\n";
    exit(255);
  }

  // Create a process group for us, then write a pid file.
  if (daemonise && daemon(1, 1))
  {
    logme()
      << "ERROR: failed to daemonise: " << strerror(errno)
      << " (error code " << errno << ")\n";
    exit(255);
  }

  params.stopfile = Filename(params.statedir, "stop");
  Filename::remove(params.stopfile, false, true);

  Filename pidfile(params.statedir, "pid");
  FILE *f = fopen(pidfile, "w");
  fprintf(f, "%ld", (long) getpgid(0));
  fclose(f);

  // Start the main process.  If it dies, restart it.
  pid_t mainproc;
  while (true)
  {
    params.crashobj[0] = 0;
    params.blacklist.clear();
    if (params.blacklistFile.exists() && params.blacklistFile.isReadable())
    {
      std::string line;
      std::ifstream f(params.blacklistFile);
      do
      {
	getline(f, line);
        if (line.empty())
	  continue;

	StringList s = StringOps::split(line, ' ', 0, 2, 0, -1);
	params.blacklist[s[1]] = Time::ValueType(atof(s[0].c_str())*1e9);
      } while (f);

      VisDQMImageServer::BlackList::iterator i = params.blacklist.begin();
      VisDQMImageServer::BlackList::iterator e = params.blacklist.end();
      for ( ; i != e; ++i)
        logme()
	  << "INFO: black-listed '" << i->first << "' since "
	  << i->second.format(true, "%Y-%m-%d %H:%M:%S.")
	  << i->second.nanoformat()
	  << " (unix time " << i->second << ")\n";
    }

    if (!daemonise || (mainproc = fork()) == 0)
    {
      // Re-capture signals from ROOT after ROOT has initialised.
      ROOT::GetROOT();
      for (int sig = 1; sig < _NSIG; ++sig)
	if (sig == SIGPIPE)
	  Signal::ignore(sig);
	else
	  Signal::revert(sig);

      Signal::handleFatal(argv_[0], IOFD_INVALID, 0, 0, FATAL_OPTS);

      // Start serving.
      runServer(params);
      exit(0);
    }

    else if (mainproc < 0)
      logme()
	<< "WARNING: failed to fork main process: "
	<< strerror(errno) << std::endl;

    else
    {
      int status = 0;
      wait(&status);

      // If the child exits, restart it.
      if (WIFEXITED(status) && WEXITSTATUS(status))
	logme()
	  << "WARNING: " << argv_[0] << "[" << mainproc
	  << "] exited with code " << WEXITSTATUS(status)
	  << ", restarting in 3 seconds\n";

      // If it died for a signal other than SIGKILL, restart it.
      else if (WIFSIGNALED(status)
	       && WTERMSIG(status) != SIGKILL
	       && WTERMSIG(status) != SIGINT
	       && WTERMSIG(status) != SIGTERM)
      {
	logme()
	  << "WARNING: " << argv_[0] << "[" << mainproc
	  << "] died of signal " << WTERMSIG(status) << " ("
	  << Signal::name (WTERMSIG(status)) << ")"
	  << " while rendering '" << params.crashobj << "'"
	  << ", restarting in 3 seconds\n";

	if (params.crashobj[0])
	  params.blacklist[params.crashobj] = Time::current();

	std::ofstream f(params.blacklistFile);
	VisDQMImageServer::BlackList::iterator i = params.blacklist.begin();
	VisDQMImageServer::BlackList::iterator e = params.blacklist.end();
	for ( ; i != e; ++i)
	  f << i->second << " " << i->first << "\n";
      }
      else
	// Signal to exit from.
	exit (0);
    }

    TimeInfo::sleep (3);
  }

  // Unreachable.
  return 0;
}

bool
VisDQMRenderDaemon::parseOption(int & /* i */, int /* argc */, char ** /* argv */)
{
  return false;
}

std::string
VisDQMRenderDaemon::showOptions(void)
{
  return std::string();
}

std::ostream &
VisDQMRenderDaemon::logme(void)
{
  return std::cerr
    << Time::current().format(true, "%Y-%m-%d %H:%M:%S")
    << " " << argv_[0] << "[" << getpid() << "]: ";
}

// Always abort on assertion failure.
char
VisDQMRenderDaemon::onAssertFail(const char *message)
{
  std::cout.flush();
  fflush(stdout);
  std::cerr.flush();
  fflush(stderr);
  std::cerr << message << "ABORTING\n";
  return 'a';
}
